from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.test import TestCase

from .utils import TestUser, is_main_site, is_paid_user
from ..menu import MenuBase
from ..templatetags.menu_generator import get_menu


class MenuTestCase(TestCase):
    """
    Menu Test
    """

    def setUp(self):
        """
        Setup the test.
        """
        self.request = HttpRequest()
        self.request.path = '/'
        self.menu = MenuBase()

    def test_template_tag(self):
        self.request.user = TestUser(authenticated=True)
        ctx = {
            'request': self.request
        }
        nav = get_menu(ctx, 'NAV_MENU')
        # While NAV_MENU is present on app1 menus and app2 menus and settings it will get 7 items
        self.assertEqual(len(nav), 7)
        nav = get_menu(ctx, 'SECOND_MENU')
        self.assertEqual(len(nav), 1)
        nav = get_menu(ctx, 'NONEXISTENT_MENU')
        self.assertEqual(len(nav), 1)
        self.assertEqual(nav[0]['name'], 'MENU NOT FOUND')

    def test_has_attr(self):
        self.assertFalse(self.menu._has_attr({}, 'name'))
        self.assertFalse(self.menu._has_attr({'name': ''}, 'name'))
        self.assertTrue(self.menu._has_attr({'name': 'Some Name'}, 'name'))
        self.assertFalse(self.menu._has_attr({}, 'url'))
        self.assertFalse(self.menu._has_attr({'url': ''}, 'url'))
        self.assertTrue(self.menu._has_attr({'url': '/'}, 'url'))
        self.assertTrue(self.menu._has_attr({'url': '/foo/bar'}, 'url'))

    def test_get_url(self):
        self.assertEqual(self.menu._get_url({}), '')
        self.assertEqual(self.menu._get_url({'url': '/'}), '/')
        self.assertEqual(self.menu._get_url({'url': '/foo/bar'}), '/foo/bar')
        self.assertEqual(self.menu._get_url({'url': 'named_url'}), '/named-url')
        params_dict = {"viewname": 'named_with_params', "kwargs": {'pk': 1}}
        self.assertEqual(self.menu._get_url({'url': params_dict}), '/named-with-params/1/')

    def test_state_anonymous_user(self):
        self.request.user = TestUser()
        self.menu.save_user_state(self.request)
        self.assertFalse(self.menu.request.user.is_authenticated)

    def test_state_regular_user(self):
        self.request.user = TestUser(authenticated=True)
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu.request.user.is_authenticated)
        self.assertFalse(self.menu.request.user.is_staff)
        self.assertFalse(self.menu.request.user.is_superuser)

    def test_state_staff(self):
        self.request.user = TestUser(authenticated=True, staff=True)
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu.request.user.is_staff)

    def test_state_superuser(self):
        self.request.user = TestUser(authenticated=True, superuser=True)
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu.request.user.is_superuser)

    def test_menu_is_validated_for_non_list_or_dict_validators(self):
        menu_dict = {
            "validators": "menu_generator.tests.utils.is_main_site",
        }
        with self.assertRaises(ImproperlyConfigured):
            self.menu._is_validated(menu_dict)

    def test_menu_is_validated_for_dict_validators(self):
        menu_dict = {
            "validators": ("menu_generator.tests.utils.is_main_site",),
        }
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_dotted_notation(self):
        menu_dict = {
            "validators": ["menu_generator.tests.utils.is_main_site"],
        }
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_invalid_dotted_notation(self):
        menu_dict = {
            "validators": ["foobar.hello"],
        }
        with self.assertRaises(ImportError):
            self.menu._is_validated(menu_dict)

    def test_menu_is_validated_callables(self):
        menu_dict = {
            "validators": [is_main_site],
        }
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_without_validators(self):
        menu_dict = {
            # no validators
        }
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_for_invalid_tuple_validator(self):
        # Test case when the coder doesn't pass parameters to the validators via tuple
        menu_dict = {
            "validators": [("menu_generator.tests.utils.is_main_site",)]
        }
        self.assertRaises(ImproperlyConfigured, self.menu._is_validated, menu_dict)

    def test_menu_is_validated_for_valid_tuple_validator(self):
        # Test case with correct parameters.
        menu_dict = {
            "validators": [
                ("menu_generator.tests.utils.validator_with_parameters", "param1", "param2",),
                "menu_generator.tests.utils.is_main_site"
            ]
        }
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_for_authenticated_users(self):
        menu_dict = {
            "validators": ["menu_generator.validators.is_authenticated"],
        }
        self.request.user = TestUser(authenticated=True)
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_for_unauthenticated_users(self):
        menu_dict = {
            "validators": ["menu_generator.validators.is_anonymous"],
        }
        self.request.user = TestUser()
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_for_superusers(self):
        menu_dict = {
            "validators": ["menu_generator.validators.is_superuser"],
        }
        self.request.user = TestUser(authenticated=True, superuser=True)
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_for_staff(self):
        menu_dict = {
            "validators": ["menu_generator.validators.is_staff"],
        }
        self.request.user = TestUser(authenticated=True, staff=True)
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_for_user_with_permission(self):
        menu_dict = {
            "validators": [
                ("menu_generator.validators.user_has_permission", "test_permission",)
            ]
        }
        self.request.user = TestUser(authenticated=True)
        self.menu.save_user_state(self.request)
        self.assertFalse(self.menu._is_validated(menu_dict))
        # Let's add the permission
        self.request.user.add_perm("test_permission")
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_is_validated_for_non_user_specific_conditions(self):
        menu_dict = {
            "validators": ["menu_generator.tests.utils.is_main_site"],
        }
        self.request.user = TestUser()
        self.menu.save_user_state(self.request)
        self.assertTrue(self.menu._is_validated(menu_dict))

    def test_menu_multiple_validators(self):
        menu_dict = {
            "validators": ["menu_generator.validators.is_authenticated", "menu_generator.validators.is_staff"],
        }
        self.request.user = TestUser(authenticated=True)
        self.menu.save_user_state(self.request)
        self.assertFalse(self.menu._is_validated(menu_dict))

    def test_generate_menu_submenu_attribute_inheritance(self):
        self.request.user = TestUser(staff=True, authenticated=True, happy=True)
        self.menu.save_user_state(self.request)
        list_dict = [
            {  # Menu item -- is_saff validator will be applied to the child node
                "name": "parent1",
                "url": "/user/account/",
                "validators": ["menu_generator.validators.is_staff"],
                "submenu": [
                    {
                        "name": "child1",
                        "url": '/user/account/profile/',
                    },
                ],
            },
            {  # Menu item -- is_saff validator will be applied to the child node
                "name": "parent2",
                "url": "/user/settings/",
                "validators": ["menu_generator.validators.is_authenticated"],
                "submenu": [
                    {
                        "name": "child1",
                        "url": '/user/settings/happy/',
                        "validators": [
                            "menu_generator.tests.utils.is_user_happy",
                        ],
                    },
                    {
                        "name": "child2",
                        "url": '/user/settings/paid/',
                        "validators": [
                            is_paid_user,
                        ],
                    },
                ],
            },
        ]
        nav = self.menu.generate_menu(list_dict)
        self.assertEqual(len(nav), 2)

        self.assertTrue('menu_generator.validators.is_staff' in nav[0]['validators'])
        self.assertTrue('menu_generator.validators.is_staff' in nav[0]['submenu'][0]['validators'])

        self.assertTrue('menu_generator.validators.is_authenticated' in nav[1]['validators'])
        self.assertTrue('menu_generator.validators.is_authenticated' in nav[1]['submenu'][0]['validators'])
        self.assertTrue('menu_generator.validators.is_authenticated' in nav[1]['submenu'][1]['validators'])

        self.assertTrue('menu_generator.tests.utils.is_user_happy' in nav[1]['submenu'][0]['validators'])

        self.assertTrue(is_paid_user in nav[1]['submenu'][1]['validators'])

    def test_generate_menu_no_visible_submenu(self):
        self.request.user = TestUser(authenticated=True)
        self.menu.save_user_state(self.request)
        list_dict = [
            {  # Menu item -- is_saff validator will be applied to the child node
                "name": "parent1",
                "url": "/user/account/",
                "validators": ["menu_generator.validators.is_authenticated"],
                "submenu": [
                    {
                        "name": "child1",
                        "url": '/user/account/profile/',
                        "validators": ["menu_generator.validators.is_staff"]  # submenu no visible
                    },
                ],
            }
        ]
        nav = self.menu.generate_menu(list_dict)
        self.assertEqual(len(nav), 1)
        self.assertIsNone(nav[0]["submenu"])

    def test_generate_menu_selected_related_urls_simple(self):
        self.request.user = TestUser(authenticated=True)
        self.request.path = "/persons/1/"
        self.menu.save_user_state(self.request)
        list_dict = [
            {
                "name": "parent1",
                "url": "/user/account/",
                "related_urls": ["/persons/", "/groups/"],
            }
        ]
        nav = self.menu.generate_menu(list_dict)

        self.assertEqual(len(nav), 1)
        self.assertEqual(nav[0]["selected"], True)

    def test_generate_menu_selected_related_urls_submenu(self):
        self.request.user = TestUser(authenticated=True)
        self.request.path = "/persons/1/"
        self.menu.save_user_state(self.request)
        list_dict = [
            {
                "name": "parent1",
                "url": "/user/account/",
                "submenu": [
                    {
                        "name": "child1",
                        "url": '/user/account/profile/',
                        "related_urls": ["/persons/"]
                    },
                    {
                        "name": "child2",
                        "url": 'named_url',
                        "related_urls": ["/groups/"]
                    },
                ],
            }
        ]
        nav = self.menu.generate_menu(list_dict)

        self.assertEqual(len(nav), 1)
        self.assertEqual(nav[0]["selected"], True)
        self.assertEqual(nav[0]["submenu"][0]["selected"], True)
        self.assertEqual(nav[0]["submenu"][1]["selected"], False)



DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    },
}
SECRET_KEY = "r4dy"
INSTALLED_APPS = [
    'menu_generator',
    'menu_generator.tests.test_apps.app1.apps.MyAppConfig',
    'menu_generator.tests.test_apps.app2',
    'menu_generator.tests.test_apps.app3'
]
MIDDLEWARE_CLASSES = []
ROOT_URLCONF = 'menu_generator.tests.urls'

NAV_MENU = [
    {
        "name": "Main",
        "url": "/",
    },
    {
        "name": "Account",
        "url": "/account",
        "validators": ["menu_generator.validators.is_authenticated", ],
        "submenu": [
            {
                "name": "Profile",
                "url": '/account/profile/',
            },
        ],
    },
    {
        "name": "Create User",
        "url": "users:create"
    },
    {
        "name": "Update User",
        "url": {"viewname": "users:update", "kwargs": {'pk': 1}}
    }
]



from django.conf.urls import url

urlpatterns = [
    url('', lambda: 'foo'),
    url('named-url', lambda: 'foo', name='named_url'),
    url('named-with-params/(?P<pk>\d+)/', lambda: 'foo', name='named_with_params')
]



def is_superuser(request):
    """
    Returns True if request.user is superuser else returns False
    """
    return is_authenticated(request) and request.user.is_superuser


def is_staff(request):
    """
    Returns True if request.user is staff else returns False
    """
    return is_authenticated(request) and request.user.is_staff


def is_authenticated(request):
    """
    Returns True if request.user authenticated else returns False
    """
    return request.user.is_authenticated


def is_anonymous(request):
    """
    Returns True if request.user is not authenticated else returns False
    """
    return not request.user.is_authenticated


def user_has_permission(request, permission):
    """
    Returns True if request.user has the permission else returns False
    :param request: HttpRequest
    :param permission: Permission to be searched
    """
    return request.user.has_perm(permission)



def is_user_happy(request):
    return request.user.is_auth and request.user.is_happy  # pragma: no cover


def is_paid_user(request):
    return request.user.is_auth and request.user.is_happy  # pragma: no cover


def is_main_site(request):
    """
    Non-User condition.
    """
    return True  # pragma: no covere


def validator_with_parameters(request, param1, param2):
    return True


class TestUser(object):
    """
    Test User Object.
    """
    is_auth = False
    is_staff = False
    is_superuser = False
    is_happy = False
    is_paid = False
    permissions = []

    def __init__(self, staff=False, superuser=False, authenticated=False, happy=False, paid=False):
        self.is_auth = authenticated
        self.is_staff = authenticated and staff
        self.is_superuser = authenticated and superuser
        self.is_happy = authenticated and happy
        self.is_paid = authenticated and paid
        self.permissions = []

    @property
    def is_authenticated(self):
        return self.is_auth

    def add_perm(self, permission):
        """
        Method for add a permission to test user
        :param permission: Permission to be added
        """
        self.permissions.append(permission)

    def has_perm(self, permission):
        """
        Method for checking if a test user has a permission
        :param permission: Permission to be checked
        :return: Boolean indicating if a test user has a permission or not
        """
        return permission in self.permissions



MENUS = {
    'NAV_MENU': [
        {
            "name": "App1 Feature",
            "url": "/app1-feature"
        }
    ],
    'SECOND_MENU': [
        {
            "name": "Second Menu Feature",
            "url": "named_url"
        }
    ]
}



from django.conf import settings
from ..utils import get_callable, clean_app_config

MENU_DICT = ".menus.MENUS"


def get_menu_from_apps(menu_name):
    """
    Returns a consumable menulist for a given menu_name found in each menus.py file (if exists) on each app on
    INSTALLED_APPS
    :param menu_name: String, name of the menu to be found
    :return: Consumable menu list
    """
    installed_apps = getattr(settings, "INSTALLED_APPS", [])
    menu_list = []
    for app in installed_apps:
        cleaned_app = clean_app_config(app)
        try:
            all_menus_dict = get_callable(cleaned_app + MENU_DICT)
        except ImportError:
            all_menus_dict = None
        except AttributeError:
            all_menus_dict = None
        if all_menus_dict:
            menu_list += all_menus_dict.get(menu_name, [])
    return menu_list



import copy

import django
from django.core.exceptions import ImproperlyConfigured
from .utils import get_callable, parse_url

if django.VERSION >= (1, 10):  # pragma: no cover
    from django.urls import reverse, NoReverseMatch
else:
    from django.core.urlresolvers import reverse, NoReverseMatch


class MenuBase(object):
    """
    Base class that generates menu list.
    """

    def __init__(self):
        self.path = ''
        self.request = None

    def save_user_state(self, request):
        """
        Given a request object, store the current user attributes
        :param request: HttpRequest
        """
        self.request = request
        self.path = request.path

    def _is_validated(self, item_dict):
        """
        Given a menu item dictionary, it returns true if the user passes all the validator's conditions, it means,
        if the user passes all the conditions, the user can see the menu
        """
        validators = item_dict.get('validators')
        if not validators:
            return True

        if not isinstance(validators, (list, tuple)):
            raise ImproperlyConfigured("validators must be a list")

        result_validations = []
        for validator in validators:
            if isinstance(validator, tuple):
                if len(validator) <= 1:
                    raise ImproperlyConfigured("You are passing a tuple validator without args %s" % str(validator))
                func = get_callable(validator[0])
                # Using a python slice to get all items after the first to build function args
                args = validator[1:]
                # Pass the request as first arg by default
                result_validations.append(func(self.request, *args))
            else:
                func = get_callable(validator)
                result_validations.append(func(self.request))  # pragma: no cover
        return all(result_validations)

    def _has_attr(self, item_dict, attr):
        """
        Given a menu item dictionary, it returns true if an attr is set.
        """
        if item_dict.get(attr, False):
            return True
        return False

    def _get_icon(self, parent_dict):
        """
        Given a menu item dictionary, this returns an icon class if one exist, or
        returns an empty string.
        """
        return parent_dict.get('icon_class', '')

    def _get_url(self, item_dict):
        """
        Given a menu item dictionary, it returns the URL or an empty string.
        """
        url = item_dict.get('url', '')
        return parse_url(url)

    def _get_related_urls(self, item_dict):
        """
        Given a menu item dictionary, it returns the relateds URLs or an empty list.
        """
        related_urls = item_dict.get('related_urls', [])
        return [parse_url(url) for url in related_urls]

    def _is_selected(self, item_dict):
        """
        Given a menu item dictionary, it returns true if `url` is on path,
        unless the item is marked as a root, in which case returns true if `url` is part of path.

        If related URLS are given, it also returns true if one of the related URLS is part of path.
        """
        url = self._get_url(item_dict)
        if self._is_root(item_dict) and url in self.path:
            return True
        elif url == self.path:
            return True
        else:
            # Go through all related URLs and test 
            for related_url in self._get_related_urls(item_dict):
                if related_url in self.path:
                    return True
        return False

    def _is_root(self, item_dict):
        """
        Given a menu item dictionary, it returns true if item is marked as a `root`.
        """
        return item_dict.get('root', False)

    def _process_breadcrums(self, menu_list):
        """
        Given a menu list, it marks the items on the current path as selected, which
        can be used as breadcrumbs
        """
        for item in menu_list:
            if item['submenu']:
                item['selected'] = self._process_breadcrums(item['submenu'])
            if item['selected']:
                return True
        return False

    def _get_submenu_list(self, parent_dict):
        """
        Given a menu item dictionary, it returns a submenu if one exist, or
        returns None.
        """
        submenu = parent_dict.get('submenu', None)
        if submenu:
            for child_dict in submenu:
                # This does a join between the menu item validators and submenu item validators and stores it on the
                # submenu's validators
                child_dict['validators'] = list(
                    set(list(parent_dict.get('validators', [])) + list(child_dict.get('validators', [])))
                )
            submenu = self.generate_menu(submenu)
            if not submenu:
                submenu = None
        return submenu

    def _get_menu_list(self, list_dict):
        """
        A generator that returns only the visible menu items.
        """
        for item in list_dict:
            if self._has_attr(item, 'name') and self._has_attr(item, 'url'):
                if self._is_validated(item):
                    yield copy.copy(item)

    def generate_menu(self, list_dict):
        """
        Given a list of dictionaries, returns a menu list.
        """
        visible_menu = []
        for item in self._get_menu_list(list_dict):
            item['url'] = self._get_url(item)
            item['selected'] = self._is_selected(item)
            item['submenu'] = self._get_submenu_list(item)
            item['icon_class'] = self._get_icon(item)
            visible_menu.append(item)
        self._process_breadcrums(visible_menu)

        return visible_menu


class Menu(MenuBase):
    """
    Class that generates menu list.
    """

    def __call__(self, request, list_dict):
        self.save_user_state(request)
        return self.generate_menu(list_dict)


generate_menu = Menu()



from django import template
from django.conf import settings

from .utils import get_menu_from_apps
from .. import defaults
from ..menu import generate_menu

register = template.Library()


@register.simple_tag(takes_context=True)
def get_menu(context, menu_name):
    """
    Returns a consumable menu list for a given menu_name found in settings.py.
    Else it returns an empty list.

    Update, March 18 2017: Now the function get the menu list from settings and append more items if found on the
    menus.py's 'MENUS' dict.
    :param context: Template context
    :param menu_name: String, name of the menu to be found
    :return: Generated menu
    """
    menu_list = getattr(settings, menu_name, defaults.MENU_NOT_FOUND)
    menu_from_apps = get_menu_from_apps(menu_name)
    # If there isn't a menu on settings but there is menu from apps we built menu from apps
    if menu_list == defaults.MENU_NOT_FOUND and menu_from_apps:
        menu_list = menu_from_apps
    # It there is a menu on settings and also on apps we merge both menus
    elif menu_list != defaults.MENU_NOT_FOUND and menu_from_apps:
        menu_list += menu_from_apps
    return generate_menu(context['request'], menu_list)



from importlib import import_module

import django
from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

if django.VERSION >= (1, 10):  # pragma: no cover
    from django.urls import reverse, NoReverseMatch
else:
    from django.core.urlresolvers import reverse, NoReverseMatch


def get_callable(func_or_path):
    """
    Receives a dotted path or a callable, Returns a callable or None
    """
    if callable(func_or_path):
        return func_or_path

    module_name = '.'.join(func_or_path.split('.')[:-1])
    function_name = func_or_path.split('.')[-1]
    _module = import_module(module_name)
    func = getattr(_module, function_name)
    return func


def clean_app_config(app_path):
    """
    Removes the AppConfig path for this app and returns the new string
    """
    apps_names = [app.name for app in apps.get_app_configs()]
    if app_path in apps_names:
        return app_path
    else:
        app_split = app_path.split('.')
        new_app = '.'.join(app_split[:-2])
        if new_app in apps_names:
            return new_app
        else:  # pragma: no cover
            raise ImproperlyConfigured(
                "The application {0} is not in the configured apps or does".format(app_path) +
                "not have the pattern app.apps.AppConfig"
            )


def parse_url(url):
    """
    Returns concrete URL for a menu dict URL attribute.
    """
    try:
        final_url = reverse(**url) if type(url) is dict else reverse(url)
    except NoReverseMatch:
        final_url = url
    return final_url



MENUS = {
    'NAV_MENU': [
        {
            "name": "App2 Feature",
            "url": "/app2-feature"
        },
        {
            "name": "App2 Second Feature",
            "url": "/app2-second-feature"
        }
    ]
}



from django.http import HttpRequest
from django.test import TestCase

from .utils import TestUser
from ..validators import is_superuser, is_staff, is_authenticated, is_anonymous, user_has_permission


class ValidatorsTestCase(TestCase):
    """
    Validators test
    """

    def setUp(self):
        """
        Setup the test.
        """
        self.request = HttpRequest()
        self.request.path = '/'

    def test_is_superuser(self):
        self.request.user = TestUser(authenticated=True, superuser=True)
        self.assertTrue(is_superuser(self.request))

    def test_is_staff(self):
        self.request.user = TestUser(authenticated=True, staff=True)
        self.assertTrue(is_staff(self.request))

    def test_is_authenticated(self):
        self.request.user = TestUser(authenticated=True)
        self.assertTrue(is_authenticated(self.request))

    def test_is_anonymous(self):
        self.request.user = TestUser()
        self.assertTrue(is_anonymous(self.request))

    def test_user_has_permission(self):
        self.request.user = TestUser(authenticated=True)
        self.request.user.add_perm("test_permission")
        self.assertTrue(user_has_permission(self.request, "test_permission"))
